import typing as t

import attr
import click
import marshmallow
from marshmallow import fields


TYPES = {t.__name__: t for t in [int, str, bool, float]}


class TypeField(marshmallow.fields.Field):
    def __init__(self, *args, **kwargs):
        self.default = kwargs.get("default", marshmallow.missing)
        super().__init__(*args, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return TYPES[value]
        except KeyError:
            if self.default == marshmallow.missing:
                raise
            return self.default


class OptionNameField(marshmallow.fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        return [value]
        if not value.startswith("-"):
            raise marshmallow.ValidationError(
                f'{value} is an option, so must start with "-".'
            )
        return [value]


class ArgumentNameField(marshmallow.fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        return [value]


class OptionSchema(marshmallow.Schema):
    param_decls = OptionNameField(data_key="name")
    type = TypeField()
    is_flag = fields.Boolean(default=False)
    help = fields.String(default=None)
    hidden = fields.Boolean(default=False)
    required = fields.Boolean(default=False)
    nargs = fields.Integer()
    multiple = fields.Boolean()
    default = fields.Field(default=None)

    @marshmallow.post_load()
    def make_option(self, validated, partial, many):
        return click.Option(**validated)


class ArgumentSchema(marshmallow.Schema):
    param_decls = ArgumentNameField(data_key="name")
    type = TypeField()
    required = fields.Boolean()
    nargs = fields.Integer()

    @marshmallow.post_load()
    def make_argument(self, validated, partial, many):
        return click.Argument(**validated)


@attr.dataclass
class RemapParam:
    new: str
    old: str


class RemapParamSchema(marshmallow.Schema):
    new = fields.String()
    old = fields.String()

    @marshmallow.post_load()
    def make_remap(self, validated, partial, many):
        return RemapParam(**validated)


@attr.dataclass
class CommandStage:
    command: str
    remap_params: t.List[RemapParam]
    params: t.Dict[str, str]


class CommandStageSchema(marshmallow.Schema):
    command = fields.String()
    remap_params = fields.List(fields.Nested(RemapParamSchema), missing=list)
    params = fields.Dict(missing=dict)

    @marshmallow.post_load()
    def make(self, validated, partial, many):
        return CommandStage(**validated)


@attr.dataclass
class CommandTestSpec:
    invocation: t.List[str]
    input: str
    output: str


class CommandTestSpecSchema(marshmallow.Schema):
    invocation = fields.List(fields.String())
    input = fields.String()
    output = fields.String()

    @marshmallow.post_load()
    def make(self, validated, partial, many):
        return CommandTestSpec(**validated)


@attr.dataclass
class CommandSpec:
    name: str
    short_help: t.Optional[str]
    help: t.Optional[str]
    arguments: t.List[click.Argument]
    options: t.List[click.Option]
    stages: t.List[CommandStage]
    inject_values: t.List[str]
    test_specs: t.List[CommandTestSpec]
    section: str


class CommandSpecSchema(marshmallow.Schema):
    name = fields.String()
    help = fields.String(default=None, missing=None)
    short_help = fields.String(default=None, missing=None)
    arguments = fields.List(fields.Nested(ArgumentSchema), missing=list)
    options = fields.List(fields.Nested(OptionSchema), missing=list)
    stages = fields.List(fields.Nested(CommandStageSchema), data_key="stage")
    inject_values = fields.List(fields.String(), missing=list)
    test_specs = fields.List(
        fields.Nested(CommandTestSpecSchema), missing=list, data_key="test"
    )
    section = fields.String(missing=None)

    @marshmallow.post_load()
    def make(self, validated, partial, many):
        return CommandSpec(**validated)



import sys
import textwrap
import types

import docutils.nodes
import docutils.parsers.rst
import docutils.utils
import sphinx.builders.text
import sphinx.events
import sphinx.util.osutil
import sphinx.writers.text


def parse_rst(text: str) -> docutils.nodes.document:
    parser = docutils.parsers.rst.Parser()
    components = (docutils.parsers.rst.Parser,)
    settings = docutils.frontend.OptionParser(
        components=components
    ).get_default_values()
    document = docutils.utils.new_document("<rst-doc>", settings=settings)
    parser.parse(text, document)
    return document


def format_text(document: docutils.nodes.document) -> str:

    app = types.SimpleNamespace(
        srcdir=None,
        confdir=None,
        outdir=None,
        doctreedir="/",
        config=types.SimpleNamespace(
            text_newlines="native",
            text_sectionchars="=",
            text_add_secnumbers=False,
            text_secnumber_suffix=".",
        ),
        tags=set(),
        events=sphinx.events.EventManager(),
        registry=types.SimpleNamespace(
            create_translator=lambda self, something, new_builder: sphinx.writers.text.TextTranslator(
                document, new_builder
            )
        ),
    )

    builder = sphinx.builders.text.TextBuilder(app)

    translator = sphinx.writers.text.TextTranslator(document, builder)
    document.walkabout(translator)
    return translator.body


def rst2text(source: str) -> str:
    document = parse_rst(source)
    return format_text(document)



import os
import sys
import typing as t

import attr
import click

import mario
import mario.doc

from . import app
from . import config
from . import declarative
from . import plug
from . import utils


config.DEFAULTS.update(
    config.load_config(
        dir_path=os.environ.get(f"{utils.NAME}_CONFIG_DIR".upper(), None)
    )
)


class ReSTCommand(click.Command):
    """Parse help as rst."""

    def format_help_text(self, ctx, formatter):

        if self.help:
            self.help = mario.doc.rst2text(self.help)
            original_wrap_text = click.formatting.wrap_text
            click.formatting.wrap_text = lambda x, *a, **kw: x
            super().format_help_text(ctx, formatter)
            click.formatting.wrap_text = original_wrap_text
            return

        if self.short_help:
            original_help = self.help
            self.help = self.short_help
            super().format_help_text(ctx, formatter)
            self.help = original_help
            return


class SectionedFormatter(click.formatting.HelpFormatter):
    def __init__(self, *args, sections, **kwargs):
        self.sections = sections
        super().__init__(*args, **kwargs)

    def write_dl(self, rows, *args, **kwargs):

        cmd_to_section = {}
        for section_name, help_section in self.sections.items():

            for command in help_section.entries:
                cmd_to_section[command] = section_name

        sections = {}
        for subcommand, help in rows:
            sections.setdefault(
                cmd_to_section.get(subcommand, "Custom commands"), []
            ).append((subcommand, help))

        for section_name, rows in sections.items():
            if rows[0][0][0] == "-":
                # with super().section('XXX'):
                super().write_dl(rows)
            else:
                with super().section(section_name):
                    super().write_dl(rows)


class SectionedContext(click.Context):
    def __init__(self, *args, sections, **kwargs):
        self.sections = sections
        super().__init__(*args, **kwargs)

    def make_formatter(self):
        """Creates the formatter for the help and usage output."""
        return SectionedFormatter(
            sections=self.sections,
            width=self.terminal_width,
            max_width=self.max_content_width,
        )


class SectionedGroup(click.Group):
    def __init__(self, *args, sections, **kwargs):
        self.sections = sections
        super().__init__(self, *args, **kwargs)

    def make_context(self, info_name, args, parent=None, **extra):
        """This function when given an info name and arguments will kick
        off the parsing and create a new :class:`Context`.  It does not
        invoke the actual command callback though.

        :param info_name: the info name for this invokation.  Generally this
                          is the most descriptive name for the script or
                          command.  For the toplevel script it's usually
                          the name of the script, for commands below it it's
                          the name of the script.
        :param args: the arguments to parse as list of strings.
        :param parent: the parent context if available.
        :param extra: extra keyword arguments forwarded to the context
                      constructor.
        """
        for key, value in click._compat.iteritems(self.context_settings):
            if key not in extra:
                extra[key] = value
        ctx = SectionedContext(
            self, info_name=info_name, parent=parent, sections=self.sections, **extra
        )
        with ctx.scope(cleanup=False):
            self.parse_args(ctx, args)
        return ctx

    def format_commands(self, ctx, formatter):
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

            rows = []
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))

            if rows:
                formatter.write_dl(rows)


CONTEXT_SETTINGS = {"default_map": config.DEFAULTS}

doc = f"""\
Mario: Python pipelines for your shell.

GitHub: https://github.com/python-mario/mario

\b
Configuration:
  Declarative config: {config.get_config_dir() / 'config.toml'}
  Python modules: {config.get_config_dir() / 'modules/*.py'}

"""


@attr.dataclass(frozen=True)
class HelpSection:
    priority: int
    entries: t.List[str]


SECTIONS = {
    "Traversals": HelpSection(
        0, ["map", "filter", "apply", "stack", "eval", "reduce", "chain"]
    ),
    "Async traversals": HelpSection(
        1,
        [
            "async-map",
            "async-apply",
            "async-filter",
            "async-chain",
            "async-map-unordered",
        ],
    ),
}

ALIASES = app.global_registry.commands


def show(x):
    if hasattr(x, "__dict__"):
        return attr.make_class(type(x).__name__, list(vars(x).keys()))(
            **{k: show(v) for k, v in vars(x).items()}
        )
    if isinstance(x, list):
        return [show(v) for v in x]
    if isinstance(x, dict):
        return {k: show(v) for k, v in x.items()}
    return repr(x)


def cli_main(pairs, **kwargs):
    app.main(pairs, **kwargs)


def version_option(ctx, param, value):
    if not value:
        return
    click.echo("mario, version " + mario.__version__)
    sys.exit()


def build_stages(command):
    def run(ctx, **cli_params):
        out = []

        for stage in command.stages:

            mapped_stage_params = {
                remap.old.lstrip("-"): cli_params[remap.new.lstrip("-")]
                for remap in stage.remap_params
            }
            mapped_stage_params.update(stage.params)
            inject_namespace = {
                k: v for k, v in cli_params.items() if k in command.inject_values
            }
            cmd = cli.get_command(ctx, stage.command)
            out.extend(
                ctx.invoke(cmd, **mapped_stage_params, inject_values=inject_namespace)
            )
        return out

    params = command.arguments + command.options

    sections = {}
    if command.section:
        sections.setdefault(command.section, []).append(command.name)
    for k, entries in sections.items():
        SECTIONS[k] = HelpSection(100, entries)

    return ReSTCommand(
        name=command.name,
        params=params,
        callback=click.pass_context(run),
        short_help=command.short_help,
        help=command.help,
    )


COMMANDS = app.global_registry.cli_functions
for k, v in ALIASES.items():

    COMMANDS[k] = build_stages(v)


cli = SectionedGroup(
    result_callback=cli_main,
    chain=True,
    context_settings=CONTEXT_SETTINGS,
    params=[
        click.Option(
            ["--max-concurrent"], type=int, default=config.DEFAULTS["max_concurrent"]
        ),
        click.Option(
            ["--exec-before"],
            help="Python source code to be executed before any stage.",
            default=config.DEFAULTS["exec_before"],
        ),
        click.Option(
            ["--base-exec-before"],
            help="Python source code to be executed before any stage; typically set in the user config file. Combined with --exec-before value. ",
            default=config.DEFAULTS["base_exec_before"],
        ),
        click.Option(
            ["--version"],
            callback=version_option,
            is_flag=True,
            help="Show the version and exit.",
        ),
    ],
    help=doc,
    commands=COMMANDS,
    sections=SECTIONS,
)



# pylint: disable=protected-access

from __future__ import generator_stop

import collections
import os
import string
import subprocess
import sys
import textwrap
import time
import urllib

import click.testing
import hypothesis
import hypothesis.strategies as st
import pytest
from tests import config
from tests import helpers

import mario
import mario.app
import mario.cli
from mario import interpret
from mario import utils


hypothesis.settings.register_profile("ci", max_examples=1000)
hypothesis.settings.register_profile("dev", max_examples=10)
hypothesis.settings.register_profile(
    "debug", max_examples=10, verbosity=hypothesis.Verbosity.verbose
)
hypothesis.settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "default"))


def assert_exception_equal(e1, e2):
    assert type(e1) == type(e2)
    assert e1.args == e2.args


@pytest.mark.parametrize("option", ["--invented-option", "-J"])
def test_raises_on_nonexistent_option(option, runner):
    args = [option, "print"]
    in_stream = "a.b.c\n"

    result = runner.invoke(mario.cli.cli, args, input=in_stream)
    assert result.exception


def test_eval_main(capsys):
    mario.app.main([[{"name": "eval", "code": "1+1", "parameters": {}}]])
    assert capsys.readouterr().out == "2\n"


def test_eval_cli():
    assert helpers.run(["eval", "1+1"]).decode() == "2\n"


def test_stack():
    args = [sys.executable, "-m", "mario", "stack", "len(x)"]
    stdin = "1\n2\n".encode()
    output = subprocess.check_output(args, input=stdin).decode()
    assert output == "4\n"


def test_chain():
    expected = "1\n1\n1\n1\n"
    result = helpers.run(["stack", "x", "chain", "map", "len"], input=b"abc\n").decode()
    assert result == expected, (result, expected)


def test_exec_before():
    args = [
        sys.executable,
        "-m",
        "mario",
        "--exec-before",
        "from collections import Counter as c",
        "stack",
        "c(x)",
    ]
    stdin = "1\n2\n".encode()
    output = subprocess.check_output(args, input=stdin).decode()
    assert output.startswith("Counter")


def test_cli_version(runner):
    args = ["--version"]
    result = helpers.run(args).decode()
    assert result == f"mario, version {mario.__version__}\n"


def test_config_file(tmp_path):
    config_body = """
    exec_before = "from collections import Counter as C"
    """

    config_file_path = tmp_path / "config.toml"

    config_file_path.write_text(config_body)

    args = ["stack", "C(x)"]
    stdin = "1\n2\n".encode()
    env = dict(os.environ)
    env.update({f"{utils.NAME}_CONFIG_DIR".upper().encode(): str(tmp_path).encode()})
    output = helpers.run(args, input=stdin, env=env).decode()
    assert output.startswith("Counter")


def test_exec_before():
    exec_before = textwrap.dedent(
        """\
    import csv
    def func(line):
        return next(csv.reader([line]))
    """
    )

    assert (
        helpers.run(
            ["--exec-before", exec_before, "map", "func"], input=b"a,b\n"
        ).decode()
        == "['a', 'b']\n"
    )


def test_stage_exec_before():
    assert helpers.run(["eval", "--exec-before", "a=1", "a"]).decode() == "1\n"
